import direct.directbase.DirectStart
from direct.showbase.DirectObject import DirectObject
from pandac.PandaModules import *
from fantasma import Fantasma
from pacman import Pacman
from pacDot import PacDot
from sensor import Sensor
from acao import Acao
from eventos import Eventos
from direct.task import Task
from leitorEscritorArquivo import LeitorEscritorArquivo
from direct.gui.OnscreenImage import OnscreenImage
from direct.gui.OnscreenText import OnscreenText
from pandac.PandaModules import TransparencyAttrib

import math
import sys

class Mundo(DirectObject):

	numParedes = 10
	numFantasmas = 3
	nomeMapa = 'eggs/mapa.egg'
	desenhoMapa = "eggs/mapaCor.egg"
	arquivoScore = 'arquivos/scoreEnviesado.txt'
	arquivoBackup = 'arquivos/parciaisEnviesado.txt'

	def __init__(self):

		#inicializa atributos da classe
		self.listaDeObjetosDoCenario = dict()
		self.contadorDePacDots = 0
		self.contadorDeFantasmas = 0
		self.pacmanPego = False
		self.handlerLateral = []

		self.rodaLabirinto()

		#roda pacman e o posiciona
		self.posicionaPacman()
		
		#roda os fantasmas e os posiciona
		self.posicionaFantasmas()		

		#posiciona pac-Dots
		self.posicionaPacDots()
		
		self.leitor = LeitorEscritorArquivo(self.arquivoScore, self.arquivoBackup)
		self.inicializaSensoresDosFantasmas()
		self.fazLevantamentoDeScore()
		
		#instancia de teclas
		self.defineTeclas()
		
		#instancia gerenciadores e eventos de colisao
		self.criaGerenciadoresDeColisao()
		self.adicionaPacmanNoGerenciador()
		self.adicionaFantasmaNoGerenciador()		
		self.adicionaBolinhaNoGerenciador()
		self.adicionaParedesNoGerenciador()
		self.adicionaPadraoDeEventosAosGerenciadores()

		self.eventos = Eventos(self)
		self.posicionaCamera()		

		self.contadorDeTempo = ClockObject()
		
		#cria Tasks
		self.criaTaskParaMovimentarFantasma()
		self.criaTaskControladora()

	def rodaLabirinto(self):
		self.environ = loader.loadModel(Mundo.nomeMapa)
		self.environ.reparentTo(render)
		self.environ.setScale(6,6,6)
		self.desenhoMapa = loader.loadModel(Mundo.desenhoMapa)
		self.desenhoMapa.reparentTo(render)
		self.desenhoMapa.setScale(6,6,6)

	def inicializaSensoresDosFantasmas(self):
		Fantasma.listaDeSensores ['sobrePacDot'] = Sensor()
		Fantasma.listaDeSensores ['pacman'] = Sensor()
		Fantasma.listaDeSensores ['fantasma'] = Sensor()
		Fantasma.listaDeSensores ['pacDot'] = Sensor()
		Fantasma.listaDeSensores ['parede'] = Sensor()

	def fazLevantamentoDeScore(self):

		listaDeLinhas = self.leitor.leArquivo()
		for linha in listaDeLinhas:
			dados = linha.split(' ')
			score = dados.pop()
			dados2 = score.split('\n')
			dados2.pop()
			scoreF = dados2.pop()		
			acao = dados.pop()
			sensor = dados.pop()
			novaAcao = Acao(acao,score)
			sensorObj = Fantasma.listaDeSensores[sensor]
			sensorObj.adicionaNaLista(novaAcao)
		Fantasma.preencheProbabilidades()

	def criaTaskParaMovimentarFantasma(self):
		taskMgr.add(self.mexeFantasma,'mexeFantasma')

	def criaTaskControladora(self):
		taskMgr.add(self.controleDeJogo,'controleDeJogo')

	def controleDeJogo(self,task):

		if self.pacmanPego == True:
			self.encerraPartida(False, self.contadorDeTempo.getLongTime())
			return task.done

		elif self.contadorDeTempo.getLongTime() > 120:
			self.encerraPartida(False, self.contadorDeTempo.getLongTime())
			return task.done
	
		elif self.contadorDePacDots == 0:
			self.encerraPartida(True, self.contadorDeTempo.getLongTime())
			return task.done

		return task.again

	def encerraPartida(self,pacmanVenceu,tempoFinalDeJogo):

		for i in range(1,self.contadorDeFantasmas+1):
			nome = 'fantasma' + str(i)
			fantasma = self.listaDeObjetosDoCenario[nome]
			for jogada in fantasma.listaDeJogadas:
				self.atribuiPontuacao(jogada,pacmanVenceu,tempoFinalDeJogo)
				sensor = Fantasma.listaDeSensores[jogada.sensor]
				acao = sensor.listaDeAcoes[jogada.nome]
				acao.score += jogada.score

		textoGravado = ""

		for nomeDoSensor, sensor in Fantasma.listaDeSensores.iteritems():
			for nomeDaAcao, acao in sensor.listaDeAcoes.iteritems():
				textoGravado += nomeDoSensor + ' ' + nomeDaAcao + ' ' + str(acao.score) + '\n'

		self.leitor.escreveNoArquivo(textoGravado) 
		self.leitor.escreveNoBackup(textoGravado)
		
		camera.setPosHpr(0, 0, 0, 0, 0, 0)
		render.removeChildren()
		render.setLightOff() 

		if pacmanVenceu:
			OnscreenImage( image= 'pacV.png', pos=(0,0,0), scale=(1.0) )
		else:
			OnscreenImage( image= 'fantV.png', pos=(0,0,0), scale=(1.0) )	

	def atribuiPontuacao(self,jogada,pacmanVenceu, tempoFinalDeJogo):
		if pacmanVenceu:
			jogada.score = int(int(jogada.tempo)/int(tempoFinalDeJogo)*100 + int(jogada.tempo))
		else:
			jogada.score = int(int(jogada.tempo)/int(tempoFinalDeJogo)*100)

	def mexeFantasma(self,task):

		for i in range(1,self.contadorDeFantasmas+1):
			task.delayTime = 0.2
			nome = 'fantasma' + str(i)
			fantasma = self.listaDeObjetosDoCenario[nome]

			fantasma.cancelaRaio()
			fantasma.lancaRaioVisao()

			fantasma.cancelaRaiosLaterais()
			fantasma.lancaRaiosLaterais()

			#adiciona raio de visao ao handler
			base.cTrav.addCollider(fantasma.raioColFrontal, self.handler)
			#base.cTrav.showCollisions(render) 

			#organiza entrada para pegar a mais proxima
			self.handler.sortEntries()
			
			paredes = []

			for i in range (0,4):
				base.cTrav.addCollider(fantasma.raioColLateral[i], self.handlerLateral[i])
				#base.cTrav.showCollisions(render) 

			for i in range (0,4):
				self.handlerLateral[i].sortEntries()
				if self.handlerLateral[i].getNumEntries() > 0:
					paredes.insert(i, self.handlerLateral[i].getEntry(0))
	
			maisPerto = None

			if self.handler.getNumEntries() > 0:
				maisPerto = self.handler.getEntry(0)
			
			if maisPerto!= None:
				fantasma.moveFantasma(maisPerto,self.contadorDeTempo.getLongTime(),paredes)
			else:
				fantasma.mexeFantasma()
			
		return task.again

	def posicionaCamera(self):
		base.disableMouse()
		camera.setPosHpr(0, -150, 200, 0, -55, 0)

	def mostraTodasBarreirasDeColisao(self):
		render.findAllMatches('**/+CollisionNode').show()

	def adicionaPadraoDeEventosAosGerenciadores(self):
		self.pusher.addInPattern('%fn-into-%in')
		self.pusher.addAgainPattern('%fn-again-%in')
		self.eventer.addInPattern('%fn-into-%in')
		self.eventer.addInPattern('%fn-again-%in')

	def adicionaParedesNoGerenciador(self):
		for i in range(0,Mundo.numParedes+1):
			parede = self.environ.find("**/parede" + str(i))
			parede.node().setName('parede')

	def adicionaBolinhaNoGerenciador(self):
		for i in range(1,self.contadorDePacDots+1):
			nome = 'pacDot' + str(i)
			pacDot = self.listaDeObjetosDoCenario[nome]
			base.cTrav.addCollider(pacDot.esferaCol,self.eventer)
		
	def adicionaPacmanNoGerenciador(self):
		pacman = self.listaDeObjetosDoCenario['pacman']
		base.cTrav.addCollider(pacman.esferaCol, self.pusher)
		self.pusher.addCollider(pacman.esferaCol, pacman.node)
		base.cTrav.addCollider(pacman.esferaColComPacDot,self.eventer)

	def adicionaFantasmaNoGerenciador(self):
		for i in range(1,self.contadorDeFantasmas+1):
			nome = 'fantasma' + str(i)
			fantasma = self.listaDeObjetosDoCenario[nome]

			base.cTrav.addCollider(fantasma.esferaCol, self.pusher)
			self.pusher.addCollider(fantasma.esferaCol, fantasma.node)
			base.cTrav.addCollider(fantasma.esferaColPacDot,self.eventer)

	def criaGerenciadoresDeColisao(self):
		base.cTrav = CollisionTraverser()
		self.pusher = CollisionHandlerPusher()
		self.handler = CollisionHandlerQueue()
		self.eventer = CollisionHandlerEvent()
		for i in range (0,4):
			self.handlerLateral.insert(i, CollisionHandlerQueue())

	def defineTeclas(self):
		pacman = self.listaDeObjetosDoCenario['pacman']
		fantasma = self.listaDeObjetosDoCenario['fantasma1']
		self.accept('escape', sys.exit)
		self.accept('w-repeat', pacman.moveParaCima)
		self.accept('d-repeat', pacman.moveParaLadoDireito)
		self.accept('a-repeat', pacman.moveParaLadoEsquerdo)
		self.accept('x-repeat', pacman.moveParaBaixo)
		self.accept('w', pacman.moveParaCima)
		self.accept('d', pacman.moveParaLadoDireito)
		self.accept('a', pacman.moveParaLadoEsquerdo)
		self.accept('x', pacman.moveParaBaixo)
		

	def posicionaPacman(self):
		pacman = Pacman()
		pacman.node.reparentTo(self.environ)
		self.listaDeObjetosDoCenario[pacman.nome] = pacman

	def posicionaFantasmas(self):
		for i in range(1,Mundo.numFantasmas+1):
			self.contadorDeFantasmas += 1
			nome = 'fantasma' + str(i)
			posicao = pow(-2,i)
			fantasma = Fantasma(nome,posicao,-1*i)
			fantasma.node.reparentTo(self.environ)
			self.listaDeObjetosDoCenario[fantasma.nome] = fantasma

	def devolveFantasma(self,nomeFantasma):

		return self.listaDeObjetosDoCenario[nomeFantasma]

	def posicionaPacDots(self):

		for i in range(1,8):
			self.contadorDePacDots += 1
			nome = 'pacDot' + str(self.contadorDePacDots)
			pacDot = PacDot(nome,i+3,8.5)
			pacDot.node.reparentTo(self.environ)
			self.listaDeObjetosDoCenario[nome] = pacDot

		for i in range(1,8):
			self.contadorDePacDots += 1
			nome = 'pacDot' + str(self.contadorDePacDots)
			pacDot = PacDot(nome,-(i+3),8.5)
			pacDot.node.reparentTo(self.environ)
			self.listaDeObjetosDoCenario[nome] = pacDot

		for i in range(1,8):
			self.contadorDePacDots += 1
			nome = 'pacDot' + str(self.contadorDePacDots)
			pacDot = PacDot(nome,i+3,-8.5)
			pacDot.node.reparentTo(self.environ)
			self.listaDeObjetosDoCenario[nome] = pacDot
				
		for i in range(1,8):
			self.contadorDePacDots += 1
			nome = 'pacDot' + str(self.contadorDePacDots)
			pacDot = PacDot(nome,-(i+3),-8.5)
			pacDot.node.reparentTo(self.environ)
			self.listaDeObjetosDoCenario[nome] = pacDot

		for i in range(8,14):
			self.contadorDePacDots += 1
			nome = 'pacDot' + str(self.contadorDePacDots)
			pacDot = PacDot(nome,2,i-6)
			pacDot.node.reparentTo(self.environ)
			self.listaDeObjetosDoCenario[nome] = pacDot

		for i in range (0,3):
			for j in range(0,3):
				self.contadorDePacDots += 1
				nome = 'pacDot' + str(self.contadorDePacDots)
				pacDot = PacDot(nome,i+7,j-2)
				pacDot.node.reparentTo(self.environ)
				self.listaDeObjetosDoCenario[nome] = pacDot

		for i in range (0,3):
			for j in range(0,3):
				self.contadorDePacDots += 1
				nome = 'pacDot' + str(self.contadorDePacDots)
				pacDot = PacDot(nome,-(i+7),j-2)
				pacDot.node.reparentTo(self.environ)
				self.listaDeObjetosDoCenario[nome] = pacDot


mundo = Mundo()
run()
